/*
  XPLDirect.cpp 
  Created by Michael Gerlicher, September 2020.
 
*/

//#include "WProgram.h"

#include <arduino.h>
#include "XPLDirect.h"

XPLDirect::XPLDirect(Stream* device)
{
	streamPtr = device;
}

void XPLDirect::begin(char * devicename)
{
//	streamPtr->print("Device name: ");
//	streamPtr->println(devicename);
	
	_deviceName = devicename;
	_connectionStatus = 0;
	_dataRefsCount = 0;
	_commandsCount = 0;
	_allDataRefsRegistered = 0;
	_receiveBufferPosition = 0;
	_receiveBuffer[0] = 0;
}


int XPLDirect::xloop(void)
{
	
	long dec;
	int i;
	long int startTime = millis();

	if (!_allDataRefsRegistered)
	{
	

		while (millis() - startTime < XPL_RETURN_TIME_MILLIS_UNREGISTERED)
			if (streamPtr->available()) _processSerial();

		return _connectionStatus;

	}

	
	//while (millis() - startTime < XPL_RETURN_TIME_MILLIS)
//	{
		if (streamPtr->available()) _processSerial();

		  //process datarefs
		for (i = 0; i < _dataRefsCount; i++)
		{
			if (_dataRefs[i]->dataRefHandle >= 0 && (_dataRefs[i]->dataRefRWType == XPL_WRITE || _dataRefs[i]->dataRefRWType == XPL_READWRITE))
			{

				switch (_dataRefs[i]->dataRefVARType)
				{
					case XPL_DATATYPE_INT:
					{

						if (*(long int*)_dataRefs[i]->latestValue != _dataRefs[i]->lastSentIntValue
							&& millis() - _dataRefs[i]->lastUpdateTime > _dataRefs[i]->updateRate || _dataRefs[i]->forceUpdate)
						{
										
							_sendPacket(XPLCMD_DATAREFUPDATE, _dataRefs[i]->dataRefHandle, (int)*(long int*)_dataRefs[i]->latestValue);

							_dataRefs[i]->lastSentIntValue = *(long int*)_dataRefs[i]->latestValue;
							_dataRefs[i]->lastUpdateTime = millis();
							_dataRefs[i]->forceUpdate = 0;

						}

						break;
					}

				

					case XPL_DATATYPE_FLOAT:
					{


						if ((*(float*)_dataRefs[i]->latestValue != _dataRefs[i]->lastSentFloatValue
							&& (millis() - _dataRefs[i]->lastUpdateTime) > _dataRefs[i]->updateRate) || _dataRefs[i]->forceUpdate)
						{
							
							_sendPacket(XPLCMD_DATAREFUPDATE, _dataRefs[i]->dataRefHandle, *(float*)_dataRefs[i]->latestValue);
							
							_dataRefs[i]->lastSentFloatValue = *(float*)_dataRefs[i]->latestValue;

							//streamPtr->println(_dataRefs[i]->lastSentFloatValue);

							
							
							_dataRefs[i]->lastUpdateTime = millis();
							_dataRefs[i]->forceUpdate = 0;
						}

						break;
					}
				}

				
			}

		}


		// now process commands
		for (i = 0; i < _commandsCount; i++)
		{
			if (_commands[i]->commandHandle >= 0)
			{
				//streamPtr->print("latest: ");  streamPtr->print(*(int *)_commands[i]->latestValue); streamPtr->print(" last sent: "); streamPtr->println(_commands[i]->lastSentValue);

				if (*(int *)_commands[i]->latestValue != _commands[i]->lastSentValue && millis() - _commands[i]->lastUpdateTime > XPL_COMMAND_MAX_UPDATE_RATE_MILLIS)

				{

					
					if (*(int *)_commands[i]->latestValue > 0)
					{
						_sendPacket(XPLCMD_COMMANDEND, _commands[i]->commandHandle);
						
					}
					else
					{
						
						_sendPacket(XPLCMD_COMMANDSTART, _commands[i]->commandHandle);
					}

					_commands[i]->lastSentValue = *(int *)_commands[i]->latestValue;
			//		streamPtr->print("updated latest: ");  streamPtr->print(*(int *)_commands[i]->latestValue); streamPtr->print(" last sent: "); streamPtr->println(_commands[i]->lastSentValue);
				}


			}

		}
//	}

	return _connectionStatus;

}

int XPLDirect::connectionStatus()
{
	return _connectionStatus;

}

int XPLDirect::allDataRefsRegistered()
{
	return _allDataRefsRegistered;
}


int XPLDirect::registerDataRef(const char *datarefName, int rwmode, unsigned int rate, float divider,  long int *value)
{
	
	if (_dataRefsCount >= XPLDIRECT_MAXDATAREFS_ARDUINO) return -1;
	_dataRefs[_dataRefsCount] = new _dataRefStructure;

	_dataRefs[_dataRefsCount]->dataRefName = datarefName;
	_dataRefs[_dataRefsCount]->dataRefRWType = rwmode;
	_dataRefs[_dataRefsCount]->divider = divider;
	_dataRefs[_dataRefsCount]->updateRate = rate;
	_dataRefs[_dataRefsCount]->dataRefVARType = XPL_DATATYPE_INT;
	_dataRefs[_dataRefsCount]->latestValue = (void *)value;
	_dataRefs[_dataRefsCount]->lastSentIntValue = -1;   // force update on first loop
	_dataRefs[_dataRefsCount]->arrayIndex = 0;			// not used unless we are referencing an array
	_dataRefs[_dataRefsCount]->dataRefHandle = -1;      // invalid until assigned by xplane
	
	_dataRefsCount++;
	_allDataRefsRegistered = 0;

    return 0;
}

int XPLDirect::registerDataRef(const char* datarefName, int rwmode, unsigned int rate, float divider, long int* value, int index)
{

	if (_dataRefsCount >= XPLDIRECT_MAXDATAREFS_ARDUINO) return -1;
	_dataRefs[_dataRefsCount] = new _dataRefStructure;

	_dataRefs[_dataRefsCount]->dataRefName = datarefName;
	_dataRefs[_dataRefsCount]->dataRefRWType = rwmode;
	_dataRefs[_dataRefsCount]->updateRate = rate;
	_dataRefs[_dataRefsCount]->divider = divider;
	_dataRefs[_dataRefsCount]->dataRefVARType = XPL_DATATYPE_INT;		// arrays are dealt with on the XPlane plugin side
	_dataRefs[_dataRefsCount]->latestValue = (void *)value;
	_dataRefs[_dataRefsCount]->lastSentIntValue = -1;   // force update on first loop
	_dataRefs[_dataRefsCount]->arrayIndex = index;			// not used unless we are referencing an array
	_dataRefs[_dataRefsCount]->dataRefHandle = -1;      // invalid until assigned by xplane

	_dataRefsCount++;
	_allDataRefsRegistered = 0;

	return 0;
}


int XPLDirect::registerDataRef(const char* datarefName, int rwmode, unsigned int rate, float divider, float* value)
{

	if (_dataRefsCount >= XPLDIRECT_MAXDATAREFS_ARDUINO) return -1;

	_dataRefs[_dataRefsCount] = new _dataRefStructure;

	_dataRefs[_dataRefsCount]->dataRefName = datarefName;
	_dataRefs[_dataRefsCount]->dataRefRWType = rwmode;
	_dataRefs[_dataRefsCount]->dataRefVARType = XPL_DATATYPE_FLOAT;		
	_dataRefs[_dataRefsCount]->latestValue = (void *)value;
	_dataRefs[_dataRefsCount]->lastSentFloatValue = -1;   // force update on first loop
	_dataRefs[_dataRefsCount]->updateRate = rate;
	_dataRefs[_dataRefsCount]->divider = divider;
	_dataRefs[_dataRefsCount]->arrayIndex = 0;			// not used unless we are referencing an array
	_dataRefs[_dataRefsCount]->dataRefHandle = -1;      // invalid until assigned by xplane

	_dataRefsCount++;
	_allDataRefsRegistered = 0;

	return 0;
}

int XPLDirect::registerDataRef(const char* datarefName, int rwmode, unsigned int rate,  float divider, float* value, int index)
{

	if (_dataRefsCount >= XPLDIRECT_MAXDATAREFS_ARDUINO) return -1;
	_dataRefs[_dataRefsCount] = new _dataRefStructure;

	_dataRefs[_dataRefsCount]->dataRefName = datarefName;
	_dataRefs[_dataRefsCount]->dataRefRWType = rwmode;
	_dataRefs[_dataRefsCount]->dataRefVARType = XPL_DATATYPE_FLOAT;		//arrays are dealt with on the Xplane plugin side
	_dataRefs[_dataRefsCount]->latestValue = (void *)value;
	_dataRefs[_dataRefsCount]->lastSentFloatValue = -1;   // force update on first loop
	_dataRefs[_dataRefsCount]->updateRate = rate;
	_dataRefs[_dataRefsCount]->divider = divider;
	_dataRefs[_dataRefsCount]->arrayIndex = index;			// not used unless we are referencing an array
	_dataRefs[_dataRefsCount]->dataRefHandle = -1;      // invalid until assigned by xplane

	_dataRefsCount++;
	_allDataRefsRegistered = 0;

	return 0;
}

int XPLDirect::registerCommand(const char* commandName, int* value)
{

	if (_commandsCount >= XPLDIRECT_MAXCOMMANDS_ARDUINO) return -1;
	_commands[_commandsCount] = new _commandStructure;

	_commands[_commandsCount]->commandName = commandName;	
	_commands[_commandsCount]->latestValue = value;
	_commands[_commandsCount]->lastSentValue = 1;      // assume not pressed for digital switches (0 = pressed)
	_commands[_commandsCount]->commandHandle = -1;      // invalid until assigned by xplane

	_commandsCount++;
	_allDataRefsRegistered = 0;							// share this flag with the datarefs, true when everything is registered with xplane.

	return 0;
}

int XPLDirect::sendDebugMessage(char* msg)
{
	
	_sendPacket(XPLCMD_PRINTDEBUG, msg);

	return 1;
}


void XPLDirect::_sendname()
{
	if (_deviceName != NULL)
	{
		_sendPacket(XPLRESPONSE_NAME, _deviceName);
	}
	
}

void XPLDirect::_processSerial()
{
	while (streamPtr->available())
	{
		
		_receiveBuffer[_receiveBufferPosition++] = (char)streamPtr->read();
	//	streamPtr->println(rxbuffer);
	//	delay(500);

		if (_receiveBuffer[0] != XPLDIRECT_PACKETHEADER)
		{
			_receiveBuffer[0] = 0;  _receiveBufferPosition = 0;
		}

		if (_receiveBuffer[0] == XPLDIRECT_PACKETHEADER && _receiveBuffer[_receiveBufferPosition-1] == XPLDIRECT_PACKETTRAILER)
	    {
		
			_processPacket();
			_receiveBufferPosition = 0;
			_receiveBuffer[0] = 0;
		}

		
		if (_receiveBufferPosition >= XPLMAX_PACKETSIZE)   // packet size exceeded / bad packet

		{
			_receiveBufferPosition = 0;
			_receiveBuffer[0] = 0;
		}
			
	}

}

void XPLDirect::_processPacket()
{
 /*   streamPtr->print("valid frame discovered, length = ");
	streamPtr->print(rxbuffer.length());
	streamPtr->print(", here it is: ");
    streamPtr->print(rxbuffer);
	streamPtr->println();
*/
	int i;

	switch (_receiveBuffer[1])
	{
		case XPLCMD_RESET :
		{  
			_connectionStatus = false;
			break;
		}

	    case XPLCMD_SENDNAME :
		{
			_sendname();
			_connectionStatus = true;			//not considered active till you know my name

			for (i = 0; i < _dataRefsCount; i++)               // also, if name was requested reset active datarefs and commands
			{
				_dataRefs[i]->dataRefHandle = -1;					//  invalid again until assigned by Xplane
			}

			for (i = 0; i < _commandsCount; i++)
			{
				_commands[i]->commandHandle = -1;
			}

			break;
		}

		case XPLRESPONSE_DATAREF :
		{   
			// String _responseString = rxbuffer.substring(_frameHeader.length() + 2, rxbuffer.length() - _frameTrailer.length());

			for (int i = 0; i < _dataRefsCount; i++)
			{	
				//streamPtr->println(rxbuffer.substring(5, rxbuffer.length() - 1));

				if ( strncmp((char *)&_receiveBuffer[5], _dataRefs[i]->dataRefName, strlen(_dataRefs[i]->dataRefName) ) == 0 
					&& _dataRefs[i]->dataRefHandle == -1)
				{
				//	streamPtr->print("found dataref: "); streamPtr->println(rxbuffer.substring(2, 5));

					_dataRefs[i]->dataRefHandle = _getHandleFromFrame(); // parse the refhandle	
					i = _dataRefsCount;		// end checking
				}
				
			}
			
			break;
		}

		case XPLRESPONSE_COMMAND:
		{
			// String _responseString = rxbuffer.substring(_frameHeader.length() + 2, rxbuffer.length() - _frameTrailer.length());

			for (int i = 0; i < _commandsCount; i++)
			{
				//streamPtr->println(rxbuffer.substring(5, rxbuffer.length() - 1));

				if ( strncmp((char *)&_receiveBuffer[5], _commands[i]->commandName, strlen(_commands[i]->commandName) ) == 0
					&& _commands[i]->commandHandle == -1)
				{
					//	streamPtr->print("found dataref: "); streamPtr->println(rxbuffer.substring(2, 5));

					_commands[i]->commandHandle = _getHandleFromFrame();   // parse the refhandle	
					i = _commandsCount;		// end checking
				}

			}

			break;
		}

		case XPLCMD_SENDREQUEST:
		{   int packetSent = 0;
			int i = 0;

			while  (!packetSent && i < _dataRefsCount && i < XPLDIRECT_MAXDATAREFS_ARDUINO)			// send dataref registrations first
			{
				if (_dataRefs[i]->dataRefHandle < 0)
				{// some boards cant do sprintf with floats so this is a workaround
					
					
					sprintf(_sendBuffer, "%c%c%1.1i%2.2i%05i.%02i%s%c", XPLDIRECT_PACKETHEADER, XPLREQUEST_REGISTERDATAREF, _dataRefs[i]->dataRefRWType, _dataRefs[i]->arrayIndex , 
						(int)_dataRefs[i]->divider, (int)(_dataRefs[i]->divider * 100) %100 , _dataRefs[i]->dataRefName, XPLDIRECT_PACKETTRAILER);
					//streamPtr->println(sendBuffer);
					_transmitPacket();
					//streamPtr -> println("...packet sent");
					packetSent = 1;
					
				}
				i++;
			}

			i = 0;

			while (!packetSent && i < _commandsCount && i < XPLDIRECT_MAXCOMMANDS_ARDUINO)			// now send command registrations
			{
				if (_commands[i]->commandHandle < 0)
				{
					sprintf(_sendBuffer, "%c%c%s%c", XPLDIRECT_PACKETHEADER, XPLREQUEST_REGISTERCOMMAND, _commands[i]->commandName, XPLDIRECT_PACKETTRAILER);
					_transmitPacket();
					packetSent = 1;
				}
				i++;

			}

			if (!packetSent)
			{
				_allDataRefsRegistered = true;
				sprintf(_sendBuffer, "%c%c%c", XPLDIRECT_PACKETHEADER, XPLREQUEST_NOREQUESTS, XPLDIRECT_PACKETTRAILER);
				_transmitPacket();
			}

			break;

		}


		case XPLCMD_DATAREFUPDATE :
		{
			int refhandle = _getHandleFromFrame();
			
			//streamPtr->println();
			//streamPtr->print("updating refhandle ");
			//streamPtr->println(refhandle);
			
			for (int i = 0; i < _dataRefsCount; i++)
			{
				if ( _dataRefs[i]->dataRefHandle == refhandle && (_dataRefs[i]->dataRefRWType == XPL_READ || _dataRefs[i]->dataRefRWType == XPL_READWRITE) )
				{  
					if (_dataRefs[i]->dataRefVARType == XPL_DATATYPE_INT)
					{


						 _getPayloadFromFrame((long int *)_dataRefs[i]->latestValue);

						/*streamPtr->println();
						streamPtr->print(" for handle ");
						streamPtr->print(refhandle);
						streamPtr->print(" I received a value of ");
						streamPtr->print(*_dataRefs[i].latestIntValue);
						streamPtr->print(" derived from string: ");
						streamPtr->println(rxbuffer.substring(5, 12));
						streamPtr->print(" It was assigned to the DataRef with name ");
						streamPtr->println(_dataRefs[i].dataRefName); */

						
					}
					
					if (_dataRefs[i]->dataRefVARType == XPL_DATATYPE_FLOAT)
					{

						_getPayloadFromFrame((float *)_dataRefs[i]->latestValue);
						
						/*streamPtr->println();
						streamPtr->print(" for handle ");
						streamPtr->print(refhandle);
						streamPtr->print(" I received a value of ");
						streamPtr->print(*_dataRefs[i].latestIntValue);
						streamPtr->print(" derived from string: ");
						streamPtr->println(rxbuffer.substring(5, 12));
						streamPtr->print(" It was assigned to the DataRef with name ");
						streamPtr->println(_dataRefs[i].dataRefName);*/

						
					}

					i = _dataRefsCount;					//skip the rest

				}
				
			}
			
			
			break;
		}
		
		case XPLREQUEST_REFRESH:
		{
			
			for (int i = 0; i < _dataRefsCount; i++)
			{
				if (_dataRefs[i]->dataRefRWType == XPL_WRITE || _dataRefs[i]->dataRefRWType == XPL_READWRITE)
				{
					_dataRefs[i]->forceUpdate = 1;				// bypass noise and timing filters
					
				}
			}

			break;
		}

		case XPLCMD_DUMPREGISTRATIONS:
		{
			dumpRegistrations();
			break;
		}

		default:
		{
			//streamPtr->println("invalid command received");
			break;
		}

	}
}


void XPLDirect::_sendPacket(int command, int handle, int value)			// for ints
{
	sprintf(_sendBuffer, "%c%c%3.3i%07.7i%c", XPLDIRECT_PACKETHEADER, command, handle, value, XPLDIRECT_PACKETTRAILER);

	_transmitPacket();

}

void XPLDirect::_sendPacket(int command, int handle, float value)		// for floats
{
// some boards cant do sprintf with floats so this is a workaround.  What a pain.
	sprintf(_sendBuffer, "%c%c%3.3i%07i.%02i%c", 
		XPLDIRECT_PACKETHEADER, 
		command, 
		handle, 
		abs((int)value), 
		abs((int)(value * 100) % 100), 
		XPLDIRECT_PACKETTRAILER);

	if (value < 0) _sendBuffer[5] = '-';

	_transmitPacket();

}

void XPLDirect::_sendPacket(int command, int handle)			// just a command with a handle
{
  sprintf(_sendBuffer, "%c%c%3.3i%c", XPLDIRECT_PACKETHEADER, command, handle, XPLDIRECT_PACKETTRAILER);
  
  _transmitPacket();
}

void XPLDirect::_sendPacket(int command, char *str)			// for a string
{
	
	sprintf(_sendBuffer, "%c%c%s%c", XPLDIRECT_PACKETHEADER, command, str, XPLDIRECT_PACKETTRAILER);

	_transmitPacket();
}


void XPLDirect::_transmitPacket(void)
{

//	streamPtr->print("writing packet: ");
//	streamPtr->println(_sendBuffer);

	streamPtr->write(_sendBuffer);
	if (strlen(_sendBuffer) == 64) streamPtr->print(" ");			// apparantly a bug on some boards when we transmit exactly 64 bytes

}




int XPLDirect::_getHandleFromFrame()			// Assuming receive buffer is holding a good frame
{
	char holdChar;
	int handleRet;

	holdChar = _receiveBuffer[5];
	_receiveBuffer[5] = 0;
	handleRet = atoi((char*)&_receiveBuffer[2]);
	
	_receiveBuffer[5] = holdChar;
	return handleRet;

}

int XPLDirect::_getPayloadFromFrame(long int *value)			// Assuming receive buffer is holding a good frame
{
	char holdChar;
	
	holdChar = _receiveBuffer[15];
	_receiveBuffer[15] = 0;
	*value = atol((char*)&_receiveBuffer[5]);
	
	_receiveBuffer[15] = holdChar;

	return 0;

}

int XPLDirect::_getPayloadFromFrame(float *value)			// Assuming receive buffer is holding a good frame
{
	char holdChar;
	
	holdChar = _receiveBuffer[15];
	_receiveBuffer[15] = 0;
	*value = atof((char*)&_receiveBuffer[5]);

	_receiveBuffer[15] = holdChar;
	return 0;

}

void XPLDirect::dumpRegistrations(void)						// a debug function
{
	/*			// not enabled for production
	int i;
	for (i = 0; i < _dataRefsCount; i++)
	{
		streamPtr->println();
		streamPtr->print("handle: ");		streamPtr->print(_dataRefs[i]->dataRefHandle);
		streamPtr->print(" RWtype: ");		streamPtr->print(_dataRefs[i]->dataRefRWType);
		streamPtr->print(" VARtype: ");		streamPtr->print(_dataRefs[i]->dataRefVARType);
		streamPtr->print(" divider: ");		streamPtr->print(_dataRefs[i]->divider);
		streamPtr->print(" updateRate: ");	streamPtr->print(_dataRefs[i]->updateRate);
		streamPtr->print(" dataRefName: "); streamPtr->print(_dataRefs[i]->dataRefName);
		//streamPtr->print(" latestValue: ");	streamPtr->print(_dataRefs[i]->latestValue)
	
	}
	
	for (i = 0; i < _commandsCount; i++)
	{
		streamPtr->println();
		streamPtr->print("handle: ");		streamPtr->print(_commands[i]->commandHandle);
		streamPtr->print(" name: ");		streamPtr->print(_commands[i]->commandName);
		streamPtr->print(" latestValue: ");	streamPtr->print(*_commands[i]->latestValue);
		streamPtr->print(" lastSent: ");	streamPtr->print(_commands[i]->lastSentValue);
		
	}
	*/
}